package algorithms;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import chromosome.Ant;

/**
 * Max min ant system algorithm.
 * 
 * @author Ivan Vlašić
 * 
 */
public abstract class MaxMinAntSystem {

	// pheromones factor
	protected static final int ALPHA = 1;

	// heuristic factor
	protected static final int BETA = 5;

	// evaporation factor
	protected static final double RO = 0.2;

	private int size;
	private int numOfAnts;
	private int maxIter;
	private double tMin;
	private double tMax;
	private List<Ant> ants;
	private List<Integer> nodes;
	private List<Integer> neighboursAvailable;
	protected double[][] pheromones;
	protected double[][] heuristics;
	protected double[][] distances;
	

	private Ant bestGlobal;
	private Ant bestInIter;

	public MaxMinAntSystem(double[][] distances, int maxIter, int numOfAnts) {
		this.nodes = new ArrayList<>();
		this.distances = distances;
		this.size = distances.length;
		this.maxIter = maxIter;
		this.numOfAnts = numOfAnts;
		this.pheromones = new double[size][size];
		this.heuristics = calculateHeuristics();
		this.ants = new ArrayList<>();
		generateInitPopulation();
		this.neighboursAvailable = new ArrayList<>(nodes);
		calculatetMaxtMin();
		setInitPheromones();

	}

	/**
	 * Executes MaxMinAntSystem algorithm.
	 */
	public Ant runAlgorithm() {
		int stagnationCount = 0;
		for (int i = 0; i < this.maxIter; i++, stagnationCount++) {
			System.out.println("Generation number " + (i + 1));
			generatePopulation();
			for (Ant ant : ants) {
				createTour(ant);
			}

			evaluateAnts();
			bestInIter = new Ant(Collections.min(ants));
			if (bestGlobal == null || bestGlobal.getTourLength() > bestInIter.getTourLength()) {
				bestGlobal = new Ant(bestInIter);
				stagnationCount = 0;
			}
			if (stagnationCount >= this.maxIter * 0.2) {
				this.tMax = 1 / RO * bestGlobal.getTourLength();
				this.tMin = this.tMax / 2 * this.size;
				setInitPheromones();
				stagnationCount = 0;
			}

			if (i < maxIter / 2) {
				updatePheromones(bestInIter);
			} else {
				updatePheromones(bestGlobal);
			}

			evaporatePheromones();

		}
		return bestGlobal;
	}

	public abstract Integer selectNext(Integer current, List<Integer> visited, List<Integer> available);

	public abstract void calculatefitness(Ant ant);

	private void generatePopulation() {
		for (Ant ant : this.ants) {
			ant.getTour().clear();
			ant.setTourLength(0);
		}
	}

	/**
	 * Updates pheromone traces of ant's path proportional to its tour length.
	 * 
	 * @param ant
	 *            ant updating pheromon traces
	 */
	private void updatePheromones(Ant ant) {
		double delta = 1 / ant.getTourLength();

		for (int i = 0; i < ant.getTour().size() - 1; i++) {
			int first = ant.getTour().get(i);
			int second = ant.getTour().get(i + 1);
			updateTrace(delta, first, second);

		}

		Integer first = ant.getTour().get(0);
		Integer second = ant.getTour().get(ant.getTour().size() - 1);
		updateTrace(delta, second, first);

	}

	/**
	 * Updates specific trace by delta value
	 * 
	 * @param delta
	 *            ant's influence on pheromone traces
	 * @param first
	 *            starting city
	 * @param second
	 *            destination city
	 */
	private void updateTrace(double delta, int first, int second) {
		this.pheromones[first][second] += delta;
		if (this.pheromones[first][second] > tMax) {
			this.pheromones[first][second] = tMax;
		}
	}

	/**
	 * Evaporates all pheromone traces by specific value proportional to RO.
	 */
	private void evaporatePheromones() {
		for (int i = 0; i < size; i++) {
			for (int j = 0; j < size; j++) {
				this.pheromones[i][j] *= (1 - RO);
				if (this.pheromones[i][j] < tMin) {
					this.pheromones[i][j] = tMin;
				}
			}
		}

	}

	/**
	 * Evaluates tour lenghts of all ants in current population.
	 */
	private void evaluateAnts() {
		for (Ant ant : this.ants) {
			calculatefitness(ant);
		}
	}

	/**
	 * Creates tour for specific ant.
	 * 
	 * @param ant
	 *            ant to create tour for
	 */
	private void createTour(Ant ant) {
		List<Integer> available = new ArrayList<>(neighboursAvailable);
		Collections.shuffle(available);
		ant.addInTour(available.get(0));
		available.remove((Integer) ant.getTour().get(0));
		
		for (int i = 1; i < size; i++) {
			Integer currentNode = ant.getTour().get(i - 1);

			Integer node = selectNext(currentNode, ant.getTour(), available);
			ant.addInTour(node);
			available.remove((Integer) node);
		}
	}

	/**
	 * Generates initial population of ants and indexes of nodes.
	 */
	private void generateInitPopulation() {
		for (int i = 0; i < numOfAnts; i++) {
			Ant ant = new Ant();
			ants.add(ant);
		}
		for (int i = 0; i < this.size; i++) {
			nodes.add(i);
		}
	}

	/**
	 * Calculates heuristics for all paths.
	 * 
	 * @return heruistics
	 */
	private double[][] calculateHeuristics() {
		double[][] heuristics = new double[size][size];
		for (int i = 0; i < size; i++) {
			for (int j = 0; j < size; j++) {
				heuristics[i][j] = Math.pow(1 / this.distances[i][j], BETA);
			}
		}
		return heuristics;
	}

	/**
	 * Calculates max and min amounts of pheromones by running greedy antsystem
	 * algorithm.
	 */
	private void calculatetMaxtMin() {
		List<Integer> available = new ArrayList<>(neighboursAvailable);
		Ant ant = new Ant();
		ant.addInTour(nodes.get(0));
		available.remove((Integer) ant.getTour().get(0));
		
		for (int i = 1; i < this.size; i++) {
			Integer currentNode = ant.getTour().get(i - 1);
			Integer nextNode = findNextNode(currentNode, ant, available);
			ant.addInTour(nextNode);
			available.remove((Integer) nextNode);
		}
		
		calculatefitness(ant);
		tMax = 1 / (RO * ant.getTourLength());
		tMin = tMax / (2 * this.size);
	}

	private Integer findNextNode(Integer current, Ant ant, List<Integer> available) {
		int currentMin = available.get(0);
		for (int j = 0; j < distances.length; j++) {
			if (distances[current][j] < distances[current][currentMin] && current != j) {
				if (!ant.getTour().contains(j)) {
					currentMin = j;
				}
			}
		}

		return currentMin;

	}

	/**
	 * Sets pheromone traces to max value.
	 */
	private void setInitPheromones() {
		for (int i = 0; i < size; i++) {
			for (int j = 0; j < size; j++) {
				this.pheromones[i][j] = tMax;
			}
		}
	}

}
